#define BOOST_TEST_DYN_LINK
#define BOOST_TEST_MODULE ThrdPool

#include "thread_pool.h"
#include <boost/test/unit_test.hpp>
#include <memory>
#include <chrono>
#include <iostream>

using namespace Add;

BOOST_AUTO_TEST_SUITE(threadpool_test)

BOOST_AUTO_TEST_CASE(threadpool_test_init_and_get_threads_num)
{
  auto pool = std::make_unique<Threadpool>();
  BOOST_REQUIRE(pool->GetThreadsNumber() == 0);
  pool->Init(100);
  BOOST_REQUIRE(pool->GetThreadsNumber() == 100);
}

BOOST_AUTO_TEST_CASE(threadpool_test_control_sequence_0)
{
  auto pool = std::make_unique<Threadpool>();
  BOOST_REQUIRE(pool->Start() == false);
}

BOOST_AUTO_TEST_CASE(threadpool_test_control_sequence_1)
{
  auto pool = std::make_unique<Threadpool>();
  BOOST_REQUIRE(pool->Stop() == false);
}

BOOST_AUTO_TEST_CASE(threadpool_test_control_sequence_2)
{
  auto pool = std::make_unique<Threadpool>();
  pool->Init(1);
  BOOST_REQUIRE(pool->Start() == true);
  BOOST_REQUIRE(pool->Stop() == true);
}

BOOST_AUTO_TEST_CASE(threadpool_test_control_sequence_3)
{
  auto pool = std::make_unique<Threadpool>();
  pool->Init(1);
  BOOST_REQUIRE(pool->Stop() == false);
}

BOOST_AUTO_TEST_CASE(threadpool_test_control_sequence_4)
{
  auto pool = std::make_unique<Threadpool>();
  pool->Init(1);
  BOOST_REQUIRE(pool->Start() == true);
  BOOST_REQUIRE(pool->Stop() == true);
  BOOST_REQUIRE(pool->Start() == true);
}

BOOST_AUTO_TEST_CASE(threadpool_test_control_sequence_5)
{
  auto pool = std::make_unique<Threadpool>();
  pool->Init(1);
  auto temp_func = [] () {};
  BOOST_REQUIRE(pool->PostTask(temp_func) == false);
  BOOST_REQUIRE(pool->Start() == true);
  BOOST_REQUIRE(pool->PostTask(temp_func) == true);
  BOOST_REQUIRE(pool->Stop() == true);
  BOOST_REQUIRE(pool->PostTask(temp_func) == false);
}

BOOST_AUTO_TEST_CASE(threadpool_test_one_thread_multi_tasks)
{
  auto pool = std::make_unique<Threadpool>();
  pool->Init(1);
  pool->Start();
  std::atomic_int  total { 0 };;
  auto temp_fun = [&total] () {
    total++;
  };
  BOOST_REQUIRE(pool->PostTask(temp_fun) == true);
  BOOST_REQUIRE(pool->PostTask(temp_fun) == true);
  BOOST_REQUIRE(pool->PostTask(temp_fun) == true);
  pool->WaitTasks();
  BOOST_REQUIRE(total.load() == 3);
}

BOOST_AUTO_TEST_CASE(threadpool_test_multi_threads_one_task)
{
  auto pool = std::make_unique<Threadpool>();
  pool->Init(3);
  pool->Start();
  std::atomic_int  total { 0 };;
  auto temp_fun = [&total] () {
    total++;
  };
  BOOST_REQUIRE(pool->PostTask(temp_fun) == true);
  pool->WaitTasks();
  BOOST_REQUIRE(total.load() == 1);
}

BOOST_AUTO_TEST_CASE(threadpool_test_multi_threads_multi_tasks_1)
{
  auto pool = std::make_unique<Threadpool>();
  pool->Init(5);
  pool->Start();
  std::atomic_int  total { 0 };;
  auto temp_fun = [&total] () {
    total++;
  };
  BOOST_REQUIRE(pool->PostTask(temp_fun) == true);
  BOOST_REQUIRE(pool->PostTask(temp_fun) == true);
  BOOST_REQUIRE(pool->PostTask(temp_fun) == true);
  pool->WaitTasks();
  BOOST_REQUIRE(total.load() == 3);
}

BOOST_AUTO_TEST_CASE(threadpool_test_multi_threads_multi_tasks_2)
{
  auto pool = std::make_unique<Threadpool>();
  pool->Init(3);
  pool->Start();
  std::atomic_int  total { 0 };;
  auto temp_fun = [&total] () {
    total++;
  };
  BOOST_REQUIRE(pool->PostTask(temp_fun) == true);
  BOOST_REQUIRE(pool->PostTask(temp_fun) == true);
  BOOST_REQUIRE(pool->PostTask(temp_fun) == true);
  BOOST_REQUIRE(pool->PostTask(temp_fun) == true);
  BOOST_REQUIRE(pool->PostTask(temp_fun) == true);
  BOOST_REQUIRE(pool->PostTask(temp_fun) == true);
  BOOST_REQUIRE(pool->PostTask(temp_fun) == true);
  pool->WaitTasks();
  BOOST_REQUIRE(total.load() == 7);
}

BOOST_AUTO_TEST_CASE(threadpool_test_post_task_when_exiting)
{
  auto pool = std::make_unique<Threadpool>();
  pool->Init(3);
  pool->Start();
  std::atomic_int  total { 0 };;
  auto temp_fun = [&total] () {
    std::this_thread::sleep_for(std::chrono::milliseconds(10));
    total++;
  };
  BOOST_REQUIRE(pool->PostTask(temp_fun) == true);
  BOOST_REQUIRE(pool->PostTask(temp_fun) == true);
  pool->Stop();
  BOOST_REQUIRE(pool->PostTask(temp_fun) == false);
  BOOST_REQUIRE(pool->PostTask(temp_fun) == false);
  BOOST_REQUIRE(pool->PostTask(temp_fun) == false);
  BOOST_REQUIRE(pool->PostTask(temp_fun) == false);
  pool->WaitTasks();
  BOOST_REQUIRE(total.load() <= 2);
}

BOOST_AUTO_TEST_CASE(threadpool_test_wait_tasks)
{
  auto pool = std::make_unique<Threadpool>();
  pool->Init(3);
  pool->Start();
  std::atomic_int  total { 0 };;
  auto temp_fun = [&total] () {
    std::this_thread::sleep_for(std::chrono::milliseconds(10));
    total++;
  };
  BOOST_REQUIRE(pool->PostTask(temp_fun) == true);
  BOOST_REQUIRE(pool->PostTask(temp_fun) == true);
  BOOST_REQUIRE(pool->PostTask(temp_fun) == true);
  BOOST_REQUIRE(pool->PostTask(temp_fun) == true);
  BOOST_REQUIRE(pool->PostTask(temp_fun) == true);
  BOOST_REQUIRE(pool->PostTask(temp_fun) == true);
  pool->WaitTasks();
  BOOST_REQUIRE(total.load() == 6);
}

BOOST_AUTO_TEST_CASE(threadpool_test_cancel_tasks)
{
  auto pool = std::make_unique<Threadpool>();
  pool->Init(1);
  pool->Start();
  std::atomic_int  total { 0 };;
  auto temp_fun = [&total] () {
    std::this_thread::sleep_for(std::chrono::milliseconds(10));
    total++;
  };
  BOOST_REQUIRE(pool->PostTask(temp_fun) == true);
  BOOST_REQUIRE(pool->PostTask(temp_fun) == true);
  BOOST_REQUIRE(pool->PostTask(temp_fun) == true);
  BOOST_REQUIRE(pool->PostTask(temp_fun) == true);
  BOOST_REQUIRE(pool->PostTask(temp_fun) == true);
  BOOST_REQUIRE(pool->PostTask(temp_fun) == true);
  pool->CancelTasks();
  BOOST_CHECK(total < 6);
}

BOOST_AUTO_TEST_CASE(threadpool_test_get_tasks_number)
{
  auto pool = std::make_unique<Threadpool>();
  pool->Init(2);
  pool->Start();
  std::atomic_int  total { 0 };;
  auto temp_fun = [&total] () {
    std::this_thread::sleep_for(std::chrono::milliseconds(10));
    total++;
  };
  BOOST_REQUIRE(pool->GetTasksNumber() == 0);
  BOOST_REQUIRE(pool->PostTask(temp_fun) == true);
  BOOST_REQUIRE(pool->PostTask(temp_fun) == true);
  pool->WaitTasks();
  BOOST_REQUIRE(pool->GetTasksNumber() == 0);
  BOOST_REQUIRE(pool->PostTask(temp_fun) == true);
  BOOST_REQUIRE(pool->PostTask(temp_fun) == true);
  BOOST_REQUIRE(pool->PostTask(temp_fun) == true);
  BOOST_REQUIRE(pool->PostTask(temp_fun) == true);
  pool->WaitTasks();
  BOOST_REQUIRE(pool->GetTasksNumber() == 0);
  BOOST_REQUIRE(pool->PostTask(temp_fun) == true);
  BOOST_REQUIRE(pool->PostTask(temp_fun) == true);
  BOOST_REQUIRE(pool->PostTask(temp_fun) == true);
  BOOST_REQUIRE(pool->PostTask(temp_fun) == true);
  BOOST_REQUIRE(pool->PostTask(temp_fun) == true);
  BOOST_REQUIRE(pool->PostTask(temp_fun) == true);
  BOOST_REQUIRE(pool->PostTask(temp_fun) == true);
  BOOST_REQUIRE(pool->PostTask(temp_fun) == true);
  BOOST_REQUIRE(pool->PostTask(temp_fun) == true);
  pool->CancelTasks();
  BOOST_CHECK(pool->GetTasksNumber() < 9u);
}

BOOST_AUTO_TEST_CASE(threadpool_multi_init)
{
  auto pool = std::make_unique<Threadpool>();
  auto thread1 = std::thread([&pool] () {
    for (int i = 0; i < 5; i++) {
      std::this_thread::sleep_for(std::chrono::nanoseconds(1));
      pool->Init(3);
    }
  });
  auto thread2 = std::thread([&pool] () {
    for (int i = 0; i < 5; i++) {
      std::this_thread::sleep_for(std::chrono::nanoseconds(1));
      pool->Init(5);
    }
  });
  thread1.join();
  thread2.join();
  auto threadcnt = pool->GetThreadsNumber();
  auto right = (threadcnt == 3u || threadcnt == 5u);
  BOOST_REQUIRE(right == true);
}

BOOST_AUTO_TEST_CASE(threadpool_multi_start_stop)
{
  auto pool = std::make_unique<Threadpool>();
  pool->Init(5);
  auto thread1 = std::thread([&pool] () {
    for (int i = 0; i < 5; i++) {
      std::this_thread::sleep_for(std::chrono::nanoseconds(1));
      pool->Start();
    }
  });
  auto thread2 = std::thread([&pool] () {
    for (int i = 0; i < 5; i++) {
      std::this_thread::sleep_for(std::chrono::nanoseconds(1));
      pool->Stop();
    }
  });
  thread1.join();
  thread2.join();
  auto right = (pool->Start() == true || pool->Stop() == true);
  BOOST_REQUIRE(right == true);
}

BOOST_AUTO_TEST_CASE(threadpool_multi_init_start_stop)
{
  auto pool = std::make_unique<Threadpool>();
  auto thread0 = std::thread([&pool] () {
    for (int i = 0; i < 5; i++) {
      std::this_thread::sleep_for(std::chrono::nanoseconds(1));
      pool->Init(3);
    }
  });
  auto thread1 = std::thread([&pool] () {
    for (int i = 0; i < 5; i++) {
      std::this_thread::sleep_for(std::chrono::nanoseconds(1));
      pool->Start();
    }
  });
  auto thread2 = std::thread([&pool] () {
    for (int i = 0; i < 5; i++) {
      std::this_thread::sleep_for(std::chrono::nanoseconds(1));
      pool->Stop();
    }
  });
  thread0.join();
  thread1.join();
  thread2.join();
  auto threadCnt = pool->GetThreadsNumber();
  auto right = threadCnt == 3u ? (pool->Start() == true || pool->Stop() == true) : (pool->Start() == false && pool->Stop() == false);
  BOOST_REQUIRE(right == true);
}

BOOST_AUTO_TEST_SUITE_END()
