#Threadpool

##调用示例
  ```c++
  #include "thread_pool.h"
  #include <iostream>
  
  void task1() {
    // do something
  }
  
  void task2(int a, int b, int& result) {
    result = a + b;
  }
  
  int main()
  {
    Add::Threadpool pool;
    pool.Init(10);
    pool.Start();
    pool.PostTask(task1);
    int addResult = 0;
    pool.PostTask(std::bind(task2, 3, 5, std::ref(addResult)));
    pool.WaitTasks();
    std::cout << "add result: " << addResult << std::endl;
    pool.Stop();
    return 0;
  }
  ```

##接口说明
   * 构造线程池对象
  ```c++
  Threadpool();
  ```

   * 销毁线程池对象
  ```c++
  ~Threadpool();
  ```

   * 初始化线程池
   * 参数:
   *   number:  int, 线程池容纳的线程数目
   * 返回值:
   *   true: 成功初始化
   *   false: 初始化失败。可能已经初始化过了。
  ```c++
  bool Init(size_t number);
  ```
 
   * 启动线程池里各个线程
   * 启动后，各个线程会阻塞，等待任务到来。处理完任务，又会阻塞，等待其他任务
   * 参数:
   *   无
   * 返回值:
   *   true: 成功启动
   *   false: 启动失败。可能没有初始化，或者已经启动过了
  ```c++
  bool Start();
  ```

   * 停止线程池里各个线程
   * 此接口会阻塞，直到各个线程合并。如果此后需要用线程池，需要重新Start
   * 参数:
   *   无
   * 返回值:
   *   true: 成功停止
   *   false: 停止失败。可能没有初始化，或者已经停止过了
  ```c++
  bool Stop();
  ```

   * 向线程池递交任务
   * 仅仅递交任务马上返回，不会等待任务完成
   * 参数:
   *   无
   * 返回值:
   *   true: 成功递交任务
   *   false: 递交任务失败，可能线程在某处已经调用Stop了。
  bool PostTask(std::function<void()> task);

   * 等待各个任务处理结束
   * 某些使用场景，在递交任务后，可能需要各个任务完成
   * 参数:
   *   无
   * 返回值:
   *   无
  ```c++
  void WaitTasks();
  ```

   * 取消已经递交的任务
   * 任务递交后，如果尚未执行，可以用此接口取消。如果已经执行，则不会。
   * 参数:
   *   无
   * 返回值:
   *   无
  ```c++
  void CancelTasks();
  ```

   * 获取线程池线程数目
   * 参数:
   *   无
   * 返回值:
   *   size_t的数值，代表线程池初始化时，传过来的数目，即其中的线程数
  ```c++
  size_t GetThreadsNumber() const;
  ```

   * 获取线程池任务数目
   * 参数:
   *   无
   * 返回值:
   *   size_t的数值，代表递交到线程池，尚未来得及处理完的数目
  ```c++
  size_t GetTasksNumber() const;
  ```

